// SPDX-License-Identifier: GPL-2.0
// Copyright (c) 2020-2022 Intel Corporation.

#include <asm/unaligned.h>
#include <linux/acpi.h>
#include <linux/delay.h>
#include <linux/dmi.h>
#include <linux/i2c.h>
#include <linux/module.h>
#include <linux/pm_runtime.h>
#include <linux/version.h>
#include <media/v4l2-ctrls.h>
#include <media/v4l2-device.h>
#include <media/v4l2-fwnode.h>
#if IS_ENABLED(CONFIG_INTEL_SKL_INT3472)
#include <linux/clk.h>
#include <linux/gpio/consumer.h>
#elif IS_ENABLED(CONFIG_POWER_CTRL_LOGIC)
#include "power_ctrl_logic.h"
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0) && \
    IS_ENABLED(CONFIG_INTEL_VSC)
#include <linux/vsc.h>
#endif

#define OV01A1S_LINK_FREQ_400MHZ	400000000ULL
#define OV01A1S_SCLK			40000000LL
#define OV01A1S_MCLK			19200000
#define OV01A1S_DATA_LANES		1
#define OV01A1S_RGB_DEPTH		10

#define OV01A1S_REG_CHIP_ID		0x300a
#define OV01A1S_CHIP_ID			0x560141

#define OV01A1S_REG_MODE_SELECT		0x0100
#define OV01A1S_MODE_STANDBY		0x00
#define OV01A1S_MODE_STREAMING		0x01

/* vertical-timings from sensor */
#define OV01A1S_REG_VTS			0x380e
#define OV01A1S_VTS_DEF			0x0380
#define OV01A1S_VTS_MIN			0x0380
#define OV01A1S_VTS_MAX			0xffff

/* Exposure controls from sensor */
#define OV01A1S_REG_EXPOSURE		0x3501
#define OV01A1S_EXPOSURE_MIN		4
#define OV01A1S_EXPOSURE_MAX_MARGIN	8
#define OV01A1S_EXPOSURE_STEP		1

/* Analog gain controls from sensor */
#define OV01A1S_REG_ANALOG_GAIN		0x3508
#define OV01A1S_ANAL_GAIN_MIN		0x100
#define OV01A1S_ANAL_GAIN_MAX		0xffff
#define OV01A1S_ANAL_GAIN_STEP		1

/* Digital gain controls from sensor */
#define OV01A1S_REG_DIGILAL_GAIN_B	0x350A
#define OV01A1S_REG_DIGITAL_GAIN_GB	0x3510
#define OV01A1S_REG_DIGITAL_GAIN_GR	0x3513
#define OV01A1S_REG_DIGITAL_GAIN_R	0x3516
#define OV01A1S_DGTL_GAIN_MIN		0
#define OV01A1S_DGTL_GAIN_MAX		0x3ffff
#define OV01A1S_DGTL_GAIN_STEP		1
#define OV01A1S_DGTL_GAIN_DEFAULT	1024

/* Test Pattern Control */
#define OV01A1S_REG_TEST_PATTERN		0x4503
#define OV01A1S_TEST_PATTERN_ENABLE	BIT(7)
#define OV01A1S_TEST_PATTERN_BAR_SHIFT	0

enum {
	OV01A1S_LINK_FREQ_400MHZ_INDEX,
};

struct ov01a1s_reg {
	u16 address;
	u8 val;
};

struct ov01a1s_reg_list {
	u32 num_of_regs;
	const struct ov01a1s_reg *regs;
};

struct ov01a1s_link_freq_config {
	const struct ov01a1s_reg_list reg_list;
};

struct ov01a1s_mode {
	/* Frame width in pixels */
	u32 width;

	/* Frame height in pixels */
	u32 height;

	/* Horizontal timining size */
	u32 hts;

	/* Default vertical timining size */
	u32 vts_def;

	/* Min vertical timining size */
	u32 vts_min;

	/* Link frequency needed for this resolution */
	u32 link_freq_index;

	/* Sensor register settings for this resolution */
	const struct ov01a1s_reg_list reg_list;
};

static const struct ov01a1s_reg mipi_data_rate_720mbps[] = {
};

static const struct ov01a1s_reg sensor_1296x800_setting[] = {
	{0x0103, 0x01},
	{0x0302, 0x00},
	{0x0303, 0x06},
	{0x0304, 0x01},
	{0x0305, 0x90},
	{0x0306, 0x00},
	{0x0308, 0x01},
	{0x0309, 0x00},
	{0x030c, 0x01},
	{0x0322, 0x01},
	{0x0323, 0x06},
	{0x0324, 0x01},
	{0x0325, 0x68},
	{0x3002, 0xa1},
	{0x301e, 0xf0},
	{0x3022, 0x01},
	{0x3501, 0x03},
	{0x3502, 0x78},
	{0x3504, 0x0c},
	{0x3508, 0x01},
	{0x3509, 0x00},
	{0x3601, 0xc0},
	{0x3603, 0x71},
	{0x3610, 0x68},
	{0x3611, 0x86},
	{0x3640, 0x10},
	{0x3641, 0x80},
	{0x3642, 0xdc},
	{0x3646, 0x55},
	{0x3647, 0x57},
	{0x364b, 0x00},
	{0x3653, 0x10},
	{0x3655, 0x00},
	{0x3656, 0x00},
	{0x365f, 0x0f},
	{0x3661, 0x45},
	{0x3662, 0x24},
	{0x3663, 0x11},
	{0x3664, 0x07},
	{0x3709, 0x34},
	{0x370b, 0x6f},
	{0x3714, 0x22},
	{0x371b, 0x27},
	{0x371c, 0x67},
	{0x371d, 0xa7},
	{0x371e, 0xe7},
	{0x3730, 0x81},
	{0x3733, 0x10},
	{0x3734, 0x40},
	{0x3737, 0x04},
	{0x3739, 0x1c},
	{0x3767, 0x00},
	{0x376c, 0x81},
	{0x3772, 0x14},
	{0x37c2, 0x04},
	{0x37d8, 0x03},
	{0x37d9, 0x0c},
	{0x37e0, 0x00},
	{0x37e1, 0x08},
	{0x37e2, 0x10},
	{0x37e3, 0x04},
	{0x37e4, 0x04},
	{0x37e5, 0x03},
	{0x37e6, 0x04},
	{0x3800, 0x00},
	{0x3801, 0x00},
	{0x3802, 0x00},
	{0x3803, 0x00},
	{0x3804, 0x05},
	{0x3805, 0x0f},
	{0x3806, 0x03},
	{0x3807, 0x2f},
	{0x3808, 0x05},
	{0x3809, 0x00},
	{0x380a, 0x03},
	{0x380b, 0x1e},
	{0x380c, 0x05},
	{0x380d, 0xd0},
	{0x380e, 0x03},
	{0x380f, 0x80},
	{0x3810, 0x00},
	{0x3811, 0x09},
	{0x3812, 0x00},
	{0x3813, 0x08},
	{0x3814, 0x01},
	{0x3815, 0x01},
	{0x3816, 0x01},
	{0x3817, 0x01},
	{0x3820, 0xa8},
	{0x3822, 0x03},
	{0x3832, 0x28},
	{0x3833, 0x10},
	{0x3b00, 0x00},
	{0x3c80, 0x00},
	{0x3c88, 0x02},
	{0x3c8c, 0x07},
	{0x3c8d, 0x40},
	{0x3cc7, 0x80},
	{0x4000, 0xc3},
	{0x4001, 0xe0},
	{0x4003, 0x40},
	{0x4008, 0x02},
	{0x4009, 0x19},
	{0x400a, 0x01},
	{0x400b, 0x6c},
	{0x4011, 0x00},
	{0x4041, 0x00},
	{0x4300, 0xff},
	{0x4301, 0x00},
	{0x4302, 0x0f},
	{0x4503, 0x00},
	{0x4601, 0x50},
	{0x481f, 0x34},
	{0x4825, 0x33},
	{0x4837, 0x14},
	{0x4881, 0x40},
	{0x4883, 0x01},
	{0x4890, 0x00},
	{0x4901, 0x00},
	{0x4902, 0x00},
	{0x4b00, 0x2a},
	{0x4b0d, 0x00},
	{0x450a, 0x04},
	{0x450b, 0x00},
	{0x5000, 0x65},
	{0x5004, 0x00},
	{0x5080, 0x40},
	{0x5200, 0x18},
	{0x4837, 0x14},
	{0x0305, 0xf4},
	{0x0325, 0xc2},
	{0x3808, 0x05},
	{0x3809, 0x10},
	{0x380a, 0x03},
	{0x380b, 0x1e},
	{0x3810, 0x00},
	{0x3811, 0x00},
	{0x3812, 0x00},
	{0x3813, 0x09},
	{0x3820, 0x88},
	{0x373d, 0x24},
};

static const char * const ov01a1s_test_pattern_menu[] = {
	"Disabled",
	"Color Bar",
	"Top-Bottom Darker Color Bar",
	"Right-Left Darker Color Bar",
	"Color Bar type 4",
};

static const s64 link_freq_menu_items[] = {
	OV01A1S_LINK_FREQ_400MHZ,
};

static const struct ov01a1s_link_freq_config link_freq_configs[] = {
	[OV01A1S_LINK_FREQ_400MHZ_INDEX] = {
		.reg_list = {
			.num_of_regs = ARRAY_SIZE(mipi_data_rate_720mbps),
			.regs = mipi_data_rate_720mbps,
		}
	},
};

static const struct ov01a1s_mode supported_modes[] = {
	{
		.width = 1296,
		.height = 798,
		.hts = 1488,
		.vts_def = OV01A1S_VTS_DEF,
		.vts_min = OV01A1S_VTS_MIN,
		.reg_list = {
			.num_of_regs = ARRAY_SIZE(sensor_1296x800_setting),
			.regs = sensor_1296x800_setting,
		},
		.link_freq_index = OV01A1S_LINK_FREQ_400MHZ_INDEX,
	},
};

/* Fixed via MIPI DisCo since kernel v6.8 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 8, 0)
static const struct dmi_system_id skip_hwcfg_check_dmi_table[] = {
	{
		.ident = "Dell Latitude 9420",
		.matches = {
			DMI_EXACT_MATCH(DMI_SYS_VENDOR, "Dell Inc."),
			DMI_EXACT_MATCH(DMI_PRODUCT_NAME, "Latitude 9420"),
		},
	},
	{ }
};
#endif

struct ov01a1s {
	struct v4l2_subdev sd;
	struct media_pad pad;
	struct v4l2_ctrl_handler ctrl_handler;

	/* V4L2 Controls */
	struct v4l2_ctrl *link_freq;
	struct v4l2_ctrl *pixel_rate;
	struct v4l2_ctrl *vblank;
	struct v4l2_ctrl *hblank;
	struct v4l2_ctrl *exposure;
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0) && \
    IS_ENABLED(CONFIG_INTEL_VSC)
	struct v4l2_ctrl *privacy_status;

	/* VSC settings */
	struct vsc_mipi_config conf;
	struct vsc_camera_status status;
#endif

	/* Current mode */
	const struct ov01a1s_mode *cur_mode;

	/* To serialize asynchronus callbacks */
	struct mutex mutex;

	/* i2c client */
	struct i2c_client *client;

#if IS_ENABLED(CONFIG_INTEL_SKL_INT3472)
	/* GPIO for reset */
	struct gpio_desc *reset_gpio;
	/* GPIO for powerdown */
	struct gpio_desc *powerdown_gpio;
	/* Power enable */
	struct regulator *avdd;
	/* Clock provider */
	struct clk *clk;
#endif

	enum {
		OV01A1S_USE_DEFAULT = 0,
#if IS_ENABLED(CONFIG_INTEL_SKL_INT3472) || IS_ENABLED(CONFIG_POWER_CTRL_LOGIC)
		OV01A1S_USE_INT3472 = 1,
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0) && \
    IS_ENABLED(CONFIG_INTEL_VSC)
		OV01A1S_USE_INTEL_VSC = 2,
#endif
	} power_type;

	/* Streaming on/off */
	bool streaming;
};

static inline struct ov01a1s *to_ov01a1s(struct v4l2_subdev *subdev)
{
	return container_of(subdev, struct ov01a1s, sd);
}

static int ov01a1s_read_reg(struct ov01a1s *ov01a1s, u16 reg, u16 len, u32 *val)
{
	struct i2c_client *client = ov01a1s->client;
	struct i2c_msg msgs[2];
	u8 addr_buf[2];
	u8 data_buf[4] = {0};
	int ret = 0;

	if (len > sizeof(data_buf))
		return -EINVAL;

	put_unaligned_be16(reg, addr_buf);
	msgs[0].addr = client->addr;
	msgs[0].flags = 0;
	msgs[0].len = sizeof(addr_buf);
	msgs[0].buf = addr_buf;
	msgs[1].addr = client->addr;
	msgs[1].flags = I2C_M_RD;
	msgs[1].len = len;
	msgs[1].buf = &data_buf[sizeof(data_buf) - len];

	ret = i2c_transfer(client->adapter, msgs, ARRAY_SIZE(msgs));
	if (ret != ARRAY_SIZE(msgs))
		return ret < 0 ? ret : -EIO;

	*val = get_unaligned_be32(data_buf);

	return 0;
}

static int ov01a1s_write_reg(struct ov01a1s *ov01a1s, u16 reg, u16 len, u32 val)
{
	struct i2c_client *client = ov01a1s->client;
	u8 buf[6];
	int ret = 0;

	if (len > 4)
		return -EINVAL;

	put_unaligned_be16(reg, buf);
	put_unaligned_be32(val << 8 * (4 - len), buf + 2);

	ret = i2c_master_send(client, buf, len + 2);
	if (ret != len + 2)
		return ret < 0 ? ret : -EIO;

	return 0;
}

static int ov01a1s_write_reg_list(struct ov01a1s *ov01a1s,
				  const struct ov01a1s_reg_list *r_list)
{
	struct i2c_client *client = ov01a1s->client;
	unsigned int i;
	int ret = 0;

	for (i = 0; i < r_list->num_of_regs; i++) {
		ret = ov01a1s_write_reg(ov01a1s, r_list->regs[i].address, 1,
					r_list->regs[i].val);
		if (ret) {
			dev_err_ratelimited(&client->dev,
					    "write reg 0x%4.4x return err = %d",
					    r_list->regs[i].address, ret);
			return ret;
		}
	}

	return 0;
}

static int ov01a1s_update_digital_gain(struct ov01a1s *ov01a1s, u32 d_gain)
{
	struct i2c_client *client = ov01a1s->client;
	u32 real = d_gain << 6;
	int ret = 0;

	ret = ov01a1s_write_reg(ov01a1s, OV01A1S_REG_DIGILAL_GAIN_B, 3, real);
	if (ret) {
		dev_err(&client->dev, "failed to set OV01A1S_REG_DIGITAL_GAIN_B");
		return ret;
	}
	ret = ov01a1s_write_reg(ov01a1s, OV01A1S_REG_DIGITAL_GAIN_GB, 3, real);
	if (ret) {
		dev_err(&client->dev, "failed to set OV01A1S_REG_DIGITAL_GAIN_GB");
		return ret;
	}
	ret = ov01a1s_write_reg(ov01a1s, OV01A1S_REG_DIGITAL_GAIN_GR, 3, real);
	if (ret) {
		dev_err(&client->dev, "failed to set OV01A1S_REG_DIGITAL_GAIN_GR");
		return ret;
	}

	ret = ov01a1s_write_reg(ov01a1s, OV01A1S_REG_DIGITAL_GAIN_R, 3, real);
	if (ret) {
		dev_err(&client->dev, "failed to set OV01A1S_REG_DIGITAL_GAIN_R");
		return ret;
	}
	return ret;
}

static int ov01a1s_test_pattern(struct ov01a1s *ov01a1s, u32 pattern)
{
	if (pattern)
		pattern = (pattern - 1) << OV01A1S_TEST_PATTERN_BAR_SHIFT |
			  OV01A1S_TEST_PATTERN_ENABLE;

	return ov01a1s_write_reg(ov01a1s, OV01A1S_REG_TEST_PATTERN, 1, pattern);
}

static int ov01a1s_set_ctrl(struct v4l2_ctrl *ctrl)
{
	struct ov01a1s *ov01a1s = container_of(ctrl->handler,
					     struct ov01a1s, ctrl_handler);
	struct i2c_client *client = ov01a1s->client;
	s64 exposure_max;
	int ret = 0;

	/* Propagate change of current control to all related controls */
	if (ctrl->id == V4L2_CID_VBLANK) {
		/* Update max exposure while meeting expected vblanking */
		exposure_max = ov01a1s->cur_mode->height + ctrl->val -
			       OV01A1S_EXPOSURE_MAX_MARGIN;
		__v4l2_ctrl_modify_range(ov01a1s->exposure,
					 ov01a1s->exposure->minimum,
					 exposure_max, ov01a1s->exposure->step,
					 exposure_max);
	}

	/* V4L2 controls values will be applied only when power is already up */
	if (!pm_runtime_get_if_in_use(&client->dev))
		return 0;

	switch (ctrl->id) {
	case V4L2_CID_ANALOGUE_GAIN:
		ret = ov01a1s_write_reg(ov01a1s, OV01A1S_REG_ANALOG_GAIN, 2,
					ctrl->val);
		break;

	case V4L2_CID_DIGITAL_GAIN:
		ret = ov01a1s_update_digital_gain(ov01a1s, ctrl->val);
		break;

	case V4L2_CID_EXPOSURE:
		ret = ov01a1s_write_reg(ov01a1s, OV01A1S_REG_EXPOSURE, 2,
					ctrl->val);
		break;

	case V4L2_CID_VBLANK:
		ret = ov01a1s_write_reg(ov01a1s, OV01A1S_REG_VTS, 2,
					ov01a1s->cur_mode->height + ctrl->val);
		break;

	case V4L2_CID_TEST_PATTERN:
		ret = ov01a1s_test_pattern(ov01a1s, ctrl->val);
		break;

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0) && \
    IS_ENABLED(CONFIG_INTEL_VSC)
	case V4L2_CID_PRIVACY:
		dev_dbg(&client->dev, "set privacy to %d", ctrl->val);
		break;

#endif
	default:
		ret = -EINVAL;
		break;
	}

	pm_runtime_put(&client->dev);

	return ret;
}

static const struct v4l2_ctrl_ops ov01a1s_ctrl_ops = {
	.s_ctrl = ov01a1s_set_ctrl,
};

static int ov01a1s_init_controls(struct ov01a1s *ov01a1s)
{
	struct v4l2_ctrl_handler *ctrl_hdlr;
	const struct ov01a1s_mode *cur_mode;
	s64 exposure_max, h_blank;
	u32 vblank_min, vblank_max, vblank_default;
	int size;
	int ret = 0;

	ctrl_hdlr = &ov01a1s->ctrl_handler;
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0) && \
    IS_ENABLED(CONFIG_INTEL_VSC)
	ret = v4l2_ctrl_handler_init(ctrl_hdlr, 9);
#else
	ret = v4l2_ctrl_handler_init(ctrl_hdlr, 8);
#endif
	if (ret)
		return ret;

	ctrl_hdlr->lock = &ov01a1s->mutex;
	cur_mode = ov01a1s->cur_mode;
	size = ARRAY_SIZE(link_freq_menu_items);

	ov01a1s->link_freq = v4l2_ctrl_new_int_menu(ctrl_hdlr,
						    &ov01a1s_ctrl_ops,
						    V4L2_CID_LINK_FREQ,
						    size - 1, 0,
						    link_freq_menu_items);
	if (ov01a1s->link_freq)
		ov01a1s->link_freq->flags |= V4L2_CTRL_FLAG_READ_ONLY;

	ov01a1s->pixel_rate = v4l2_ctrl_new_std(ctrl_hdlr, &ov01a1s_ctrl_ops,
						V4L2_CID_PIXEL_RATE, 0,
						OV01A1S_SCLK, 1, OV01A1S_SCLK);

	vblank_min = cur_mode->vts_min - cur_mode->height;
	vblank_max = OV01A1S_VTS_MAX - cur_mode->height;
	vblank_default = cur_mode->vts_def - cur_mode->height;
	ov01a1s->vblank = v4l2_ctrl_new_std(ctrl_hdlr, &ov01a1s_ctrl_ops,
					    V4L2_CID_VBLANK, vblank_min,
					    vblank_max, 1, vblank_default);

	h_blank = cur_mode->hts - cur_mode->width;
	ov01a1s->hblank = v4l2_ctrl_new_std(ctrl_hdlr, &ov01a1s_ctrl_ops,
					    V4L2_CID_HBLANK, h_blank, h_blank,
					    1, h_blank);
	if (ov01a1s->hblank)
		ov01a1s->hblank->flags |= V4L2_CTRL_FLAG_READ_ONLY;
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0) && \
    IS_ENABLED(CONFIG_INTEL_VSC)
	ov01a1s->privacy_status = v4l2_ctrl_new_std(ctrl_hdlr,
						    &ov01a1s_ctrl_ops,
						    V4L2_CID_PRIVACY,
						    0, 1, 1, 0);
#endif

	v4l2_ctrl_new_std(ctrl_hdlr, &ov01a1s_ctrl_ops, V4L2_CID_ANALOGUE_GAIN,
			  OV01A1S_ANAL_GAIN_MIN, OV01A1S_ANAL_GAIN_MAX,
			  OV01A1S_ANAL_GAIN_STEP, OV01A1S_ANAL_GAIN_MIN);
	v4l2_ctrl_new_std(ctrl_hdlr, &ov01a1s_ctrl_ops, V4L2_CID_DIGITAL_GAIN,
			  OV01A1S_DGTL_GAIN_MIN, OV01A1S_DGTL_GAIN_MAX,
			  OV01A1S_DGTL_GAIN_STEP, OV01A1S_DGTL_GAIN_DEFAULT);
	exposure_max = cur_mode->vts_def - OV01A1S_EXPOSURE_MAX_MARGIN;
	ov01a1s->exposure = v4l2_ctrl_new_std(ctrl_hdlr, &ov01a1s_ctrl_ops,
					      V4L2_CID_EXPOSURE,
					      OV01A1S_EXPOSURE_MIN,
					      exposure_max,
					      OV01A1S_EXPOSURE_STEP,
					      exposure_max);
	v4l2_ctrl_new_std_menu_items(ctrl_hdlr, &ov01a1s_ctrl_ops,
				     V4L2_CID_TEST_PATTERN,
				     ARRAY_SIZE(ov01a1s_test_pattern_menu) - 1,
				     0, 0, ov01a1s_test_pattern_menu);
	if (ctrl_hdlr->error)
		return ctrl_hdlr->error;

	ov01a1s->sd.ctrl_handler = ctrl_hdlr;

	return 0;
}

static void ov01a1s_update_pad_format(const struct ov01a1s_mode *mode,
				      struct v4l2_mbus_framefmt *fmt)
{
	fmt->width = mode->width;
	fmt->height = mode->height;
	fmt->code = MEDIA_BUS_FMT_SGRBG10_1X10;
	fmt->field = V4L2_FIELD_NONE;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0) && \
    IS_ENABLED(CONFIG_INTEL_VSC)
static void ov01a1s_vsc_privacy_callback(void *handle,
				       enum vsc_privacy_status status)
{
	struct ov01a1s *ov01a1s = handle;

	v4l2_ctrl_s_ctrl(ov01a1s->privacy_status, !status);
}

#endif
static int ov01a1s_start_streaming(struct ov01a1s *ov01a1s)
{
	struct i2c_client *client = ov01a1s->client;
	const struct ov01a1s_reg_list *reg_list;
	int link_freq_index;
	int ret = 0;

	link_freq_index = ov01a1s->cur_mode->link_freq_index;
	reg_list = &link_freq_configs[link_freq_index].reg_list;
	ret = ov01a1s_write_reg_list(ov01a1s, reg_list);
	if (ret) {
		dev_err(&client->dev, "failed to set plls");
		return ret;
	}

	reg_list = &ov01a1s->cur_mode->reg_list;
	ret = ov01a1s_write_reg_list(ov01a1s, reg_list);
	if (ret) {
		dev_err(&client->dev, "failed to set mode");
		return ret;
	}

	ret = __v4l2_ctrl_handler_setup(ov01a1s->sd.ctrl_handler);
	if (ret)
		return ret;

	ret = ov01a1s_write_reg(ov01a1s, OV01A1S_REG_MODE_SELECT, 1,
				OV01A1S_MODE_STREAMING);
	if (ret)
		dev_err(&client->dev, "failed to start streaming");

	return ret;
}

static void ov01a1s_stop_streaming(struct ov01a1s *ov01a1s)
{
	struct i2c_client *client = ov01a1s->client;
	int ret = 0;

	ret = ov01a1s_write_reg(ov01a1s, OV01A1S_REG_MODE_SELECT, 1,
				OV01A1S_MODE_STANDBY);
	if (ret)
		dev_err(&client->dev, "failed to stop streaming");
}

static int ov01a1s_set_stream(struct v4l2_subdev *sd, int enable)
{
	struct ov01a1s *ov01a1s = to_ov01a1s(sd);
	struct i2c_client *client = ov01a1s->client;
	int ret = 0;

	if (ov01a1s->streaming == enable)
		return 0;

	mutex_lock(&ov01a1s->mutex);
	if (enable) {
		ret = pm_runtime_get_sync(&client->dev);
		if (ret < 0) {
			pm_runtime_put_noidle(&client->dev);
			mutex_unlock(&ov01a1s->mutex);
			return ret;
		}

		ret = ov01a1s_start_streaming(ov01a1s);
		if (ret) {
			enable = 0;
			ov01a1s_stop_streaming(ov01a1s);
			pm_runtime_put(&client->dev);
		}
	} else {
		ov01a1s_stop_streaming(ov01a1s);
		pm_runtime_put(&client->dev);
	}

	ov01a1s->streaming = enable;
	mutex_unlock(&ov01a1s->mutex);

	return ret;
}

static int ov01a1s_power_off(struct device *dev)
{
	struct v4l2_subdev *sd = dev_get_drvdata(dev);
	struct ov01a1s *ov01a1s = to_ov01a1s(sd);
	int ret = 0;

#if IS_ENABLED(CONFIG_INTEL_SKL_INT3472)
	if (ov01a1s->power_type == OV01A1S_USE_INT3472) {
		gpiod_set_value_cansleep(ov01a1s->reset_gpio, 1);
		gpiod_set_value_cansleep(ov01a1s->powerdown_gpio, 1);
		if (ov01a1s->avdd)
			ret = regulator_disable(ov01a1s->avdd);
		clk_disable_unprepare(ov01a1s->clk);
		msleep(20);
	}
#elif IS_ENABLED(CONFIG_POWER_CTRL_LOGIC)
	if (ov01a1s->power_type == OV01A1S_USE_INT3472)
		ret = power_ctrl_logic_set_power(0);
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0) && \
    IS_ENABLED(CONFIG_INTEL_VSC)
	if (ov01a1s->power_type == OV01A1S_USE_INTEL_VSC) {
		ret = vsc_release_camera_sensor(&ov01a1s->status);
		if (ret && ret != -EAGAIN)
			dev_err(dev, "Release VSC failed");
	}
#endif

	return ret;
}

static int ov01a1s_power_on(struct device *dev)
{
	struct v4l2_subdev *sd = dev_get_drvdata(dev);
	struct ov01a1s *ov01a1s = to_ov01a1s(sd);
	int ret = 0;

#if IS_ENABLED(CONFIG_INTEL_SKL_INT3472)
	if (ov01a1s->power_type == OV01A1S_USE_INT3472) {
		ret = clk_prepare_enable(ov01a1s->clk);
		if (ret)
			return ret;
		if (ov01a1s->avdd)
			ret = regulator_enable(ov01a1s->avdd);
		if (ret)
			return ret;
		gpiod_set_value_cansleep(ov01a1s->powerdown_gpio, 0);
		gpiod_set_value_cansleep(ov01a1s->reset_gpio, 0);
		msleep(20);
	}
#elif IS_ENABLED(CONFIG_POWER_CTRL_LOGIC)
	if (ov01a1s->power_type == OV01A1S_USE_INT3472)
		ret = power_ctrl_logic_set_power(1);
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0) && \
    IS_ENABLED(CONFIG_INTEL_VSC)
	if (ov01a1s->power_type == OV01A1S_USE_INTEL_VSC) {
		ret = vsc_acquire_camera_sensor(&ov01a1s->conf,
						ov01a1s_vsc_privacy_callback,
						ov01a1s, &ov01a1s->status);
		if (ret && ret != -EAGAIN) {
			dev_err(dev, "Acquire VSC failed");
			return ret;
		}
		__v4l2_ctrl_s_ctrl(ov01a1s->privacy_status,
				   !(ov01a1s->status.status));
	}
#endif

	return ret;
}

static int __maybe_unused ov01a1s_suspend(struct device *dev)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct v4l2_subdev *sd = i2c_get_clientdata(client);
	struct ov01a1s *ov01a1s = to_ov01a1s(sd);

	mutex_lock(&ov01a1s->mutex);
	if (ov01a1s->streaming)
		ov01a1s_stop_streaming(ov01a1s);

	mutex_unlock(&ov01a1s->mutex);

	return 0;
}

static int __maybe_unused ov01a1s_resume(struct device *dev)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct v4l2_subdev *sd = i2c_get_clientdata(client);
	struct ov01a1s *ov01a1s = to_ov01a1s(sd);
	int ret = 0;

	mutex_lock(&ov01a1s->mutex);
	if (!ov01a1s->streaming)
		goto exit;

	ret = ov01a1s_start_streaming(ov01a1s);
	if (ret) {
		ov01a1s->streaming = false;
		ov01a1s_stop_streaming(ov01a1s);
	}

exit:
	mutex_unlock(&ov01a1s->mutex);
	return ret;
}

static int ov01a1s_set_format(struct v4l2_subdev *sd,
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 0)
			      struct v4l2_subdev_pad_config *cfg,
#else
			      struct v4l2_subdev_state *sd_state,
#endif
			      struct v4l2_subdev_format *fmt)
{
	struct ov01a1s *ov01a1s = to_ov01a1s(sd);
	const struct ov01a1s_mode *mode;
	s32 vblank_def, h_blank;

	mode = v4l2_find_nearest_size(supported_modes,
				      ARRAY_SIZE(supported_modes), width,
				      height, fmt->format.width,
				      fmt->format.height);

	mutex_lock(&ov01a1s->mutex);
	ov01a1s_update_pad_format(mode, &fmt->format);
	if (fmt->which == V4L2_SUBDEV_FORMAT_TRY) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 0)
		*v4l2_subdev_get_try_format(sd, cfg, fmt->pad) = fmt->format;
#elif LINUX_VERSION_CODE < KERNEL_VERSION(6, 8, 0)
		*v4l2_subdev_get_try_format(sd, sd_state, fmt->pad) = fmt->format;
#else
		*v4l2_subdev_state_get_format(sd_state, fmt->pad) = fmt->format;
#endif
	} else {
		ov01a1s->cur_mode = mode;
		__v4l2_ctrl_s_ctrl(ov01a1s->link_freq, mode->link_freq_index);
		__v4l2_ctrl_s_ctrl_int64(ov01a1s->pixel_rate, OV01A1S_SCLK);

		/* Update limits and set FPS to default */
		vblank_def = mode->vts_def - mode->height;
		__v4l2_ctrl_modify_range(ov01a1s->vblank,
					 mode->vts_min - mode->height,
					 OV01A1S_VTS_MAX - mode->height, 1,
					 vblank_def);
		__v4l2_ctrl_s_ctrl(ov01a1s->vblank, vblank_def);
		h_blank = mode->hts - mode->width;
		__v4l2_ctrl_modify_range(ov01a1s->hblank, h_blank, h_blank, 1,
					 h_blank);
	}
	mutex_unlock(&ov01a1s->mutex);

	return 0;
}

static int ov01a1s_get_format(struct v4l2_subdev *sd,
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 0)
			      struct v4l2_subdev_pad_config *cfg,
#else
			      struct v4l2_subdev_state *sd_state,
#endif
			      struct v4l2_subdev_format *fmt)
{
	struct ov01a1s *ov01a1s = to_ov01a1s(sd);

	mutex_lock(&ov01a1s->mutex);
	if (fmt->which == V4L2_SUBDEV_FORMAT_TRY)
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 0)
		fmt->format = *v4l2_subdev_get_try_format(&ov01a1s->sd, cfg,
							  fmt->pad);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(6, 8, 0)
		fmt->format = *v4l2_subdev_get_try_format(&ov01a1s->sd,
							  sd_state, fmt->pad);
#else
		fmt->format = *v4l2_subdev_state_get_format(
							  sd_state, fmt->pad);
#endif
	else
		ov01a1s_update_pad_format(ov01a1s->cur_mode, &fmt->format);

	mutex_unlock(&ov01a1s->mutex);

	return 0;
}

static int ov01a1s_enum_mbus_code(struct v4l2_subdev *sd,
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 0)
				  struct v4l2_subdev_pad_config *cfg,
#else
				  struct v4l2_subdev_state *sd_state,
#endif
				  struct v4l2_subdev_mbus_code_enum *code)
{
	if (code->index > 0)
		return -EINVAL;

	code->code = MEDIA_BUS_FMT_SGRBG10_1X10;

	return 0;
}

static int ov01a1s_enum_frame_size(struct v4l2_subdev *sd,
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 0)
				   struct v4l2_subdev_pad_config *cfg,
#else
				   struct v4l2_subdev_state *sd_state,
#endif
				   struct v4l2_subdev_frame_size_enum *fse)
{
	if (fse->index >= ARRAY_SIZE(supported_modes))
		return -EINVAL;

	if (fse->code != MEDIA_BUS_FMT_SGRBG10_1X10)
		return -EINVAL;

	fse->min_width = supported_modes[fse->index].width;
	fse->max_width = fse->min_width;
	fse->min_height = supported_modes[fse->index].height;
	fse->max_height = fse->min_height;

	return 0;
}

static int ov01a1s_open(struct v4l2_subdev *sd, struct v4l2_subdev_fh *fh)
{
	struct ov01a1s *ov01a1s = to_ov01a1s(sd);

	mutex_lock(&ov01a1s->mutex);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 0)
	ov01a1s_update_pad_format(&supported_modes[0],
				  v4l2_subdev_get_try_format(sd, fh->pad, 0));
#elif LINUX_VERSION_CODE < KERNEL_VERSION(6, 8, 0)
	ov01a1s_update_pad_format(&supported_modes[0],
				  v4l2_subdev_get_try_format(sd, fh->state, 0));
#else
	ov01a1s_update_pad_format(&supported_modes[0],
				  v4l2_subdev_state_get_format(fh->state, 0));
#endif
	mutex_unlock(&ov01a1s->mutex);

	return 0;
}

static const struct v4l2_subdev_video_ops ov01a1s_video_ops = {
	.s_stream = ov01a1s_set_stream,
};

static const struct v4l2_subdev_pad_ops ov01a1s_pad_ops = {
	.set_fmt = ov01a1s_set_format,
	.get_fmt = ov01a1s_get_format,
	.enum_mbus_code = ov01a1s_enum_mbus_code,
	.enum_frame_size = ov01a1s_enum_frame_size,
};

static const struct v4l2_subdev_ops ov01a1s_subdev_ops = {
	.video = &ov01a1s_video_ops,
	.pad = &ov01a1s_pad_ops,
};

static const struct media_entity_operations ov01a1s_subdev_entity_ops = {
	.link_validate = v4l2_subdev_link_validate,
};

static const struct v4l2_subdev_internal_ops ov01a1s_internal_ops = {
	.open = ov01a1s_open,
};

static int ov01a1s_identify_module(struct ov01a1s *ov01a1s)
{
	struct i2c_client *client = ov01a1s->client;
	int ret;
	u32 val;

	ret = ov01a1s_read_reg(ov01a1s, OV01A1S_REG_CHIP_ID, 3, &val);
	if (ret)
		return ret;

	if (val != OV01A1S_CHIP_ID) {
		dev_err(&client->dev, "chip id mismatch: %x!=%x",
			OV01A1S_CHIP_ID, val);
		return -ENXIO;
	}

	return 0;
}

static int ov01a1s_check_hwcfg(struct device *dev)
{
	struct v4l2_fwnode_endpoint bus_cfg = {
		.bus_type = V4L2_MBUS_CSI2_DPHY
	};
	struct fwnode_handle *ep;
	struct fwnode_handle *fwnode = dev_fwnode(dev);
	unsigned int i, j;
	int ret;
	u32 ext_clk;

	if (!fwnode)
		return -ENXIO;

	ep = fwnode_graph_get_next_endpoint(fwnode, NULL);
	if (!ep)
		return -EPROBE_DEFER;

	ret = fwnode_property_read_u32(dev_fwnode(dev), "clock-frequency",
				       &ext_clk);
	if (ret) {
		dev_err(dev, "can't get clock frequency");
		return ret;
	}

	ret = v4l2_fwnode_endpoint_alloc_parse(ep, &bus_cfg);
	fwnode_handle_put(ep);
	if (ret)
		return ret;

	if (bus_cfg.bus.mipi_csi2.num_data_lanes != OV01A1S_DATA_LANES) {
		dev_err(dev, "number of CSI2 data lanes %d is not supported",
			bus_cfg.bus.mipi_csi2.num_data_lanes);
		ret = -EINVAL;
		goto out_err;
	}

	if (!bus_cfg.nr_of_link_frequencies) {
		dev_err(dev, "no link frequencies defined");
		ret = -EINVAL;
		goto out_err;
	}

	for (i = 0; i < ARRAY_SIZE(link_freq_menu_items); i++) {
		for (j = 0; j < bus_cfg.nr_of_link_frequencies; j++) {
			if (link_freq_menu_items[i] ==
				bus_cfg.link_frequencies[j])
				break;
		}

		if (j == bus_cfg.nr_of_link_frequencies) {
			dev_err(dev, "no link frequency %lld supported",
				link_freq_menu_items[i]);
			ret = -EINVAL;
			goto out_err;
		}
	}

out_err:
	v4l2_fwnode_endpoint_free(&bus_cfg);

	return ret;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 1, 0)
static int ov01a1s_remove(struct i2c_client *client)
#else
static void ov01a1s_remove(struct i2c_client *client)
#endif
{
	struct v4l2_subdev *sd = i2c_get_clientdata(client);
	struct ov01a1s *ov01a1s = to_ov01a1s(sd);

	v4l2_async_unregister_subdev(sd);
	media_entity_cleanup(&sd->entity);
	v4l2_ctrl_handler_free(sd->ctrl_handler);
	pm_runtime_disable(&client->dev);
	mutex_destroy(&ov01a1s->mutex);

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 1, 0)
	return 0;
#endif
}

#if IS_ENABLED(CONFIG_INTEL_SKL_INT3472)
static int ov01a1s_parse_gpio(struct ov01a1s *ov01a1s)
{
	struct device *dev = &ov01a1s->client->dev;

	ov01a1s->reset_gpio = devm_gpiod_get(dev, "reset", GPIOD_OUT_LOW);
	if (IS_ERR(ov01a1s->reset_gpio)) {
		dev_warn(dev, "error while getting reset gpio: %ld\n",
			 PTR_ERR(ov01a1s->reset_gpio));
		ov01a1s->reset_gpio = NULL;
		return -EPROBE_DEFER;
	}

	/* For optional, don't return or print warn if can't get it */
	ov01a1s->powerdown_gpio =
		devm_gpiod_get_optional(dev, "powerdown", GPIOD_OUT_LOW);
	if (IS_ERR(ov01a1s->powerdown_gpio)) {
		dev_dbg(dev, "no powerdown gpio: %ld\n",
			PTR_ERR(ov01a1s->powerdown_gpio));
		ov01a1s->powerdown_gpio = NULL;
	}

	ov01a1s->avdd = devm_regulator_get_optional(dev, "avdd");
	if (IS_ERR(ov01a1s->avdd)) {
		dev_dbg(dev, "no regulator avdd: %ld\n",
			PTR_ERR(ov01a1s->avdd));
		ov01a1s->avdd = NULL;
	}

	ov01a1s->clk = devm_clk_get_optional(dev, "clk");
	if (IS_ERR(ov01a1s->clk)) {
		dev_dbg(dev, "no clk: %ld\n", PTR_ERR(ov01a1s->clk));
		ov01a1s->clk = NULL;
	}

	return 0;
}
#endif

static int ov01a1s_parse_power(struct ov01a1s *ov01a1s)
{
	int ret = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0) && \
    IS_ENABLED(CONFIG_INTEL_VSC)
	ov01a1s->conf.lane_num = OV01A1S_DATA_LANES;
	/* frequency unit 100k */
	ov01a1s->conf.freq = OV01A1S_LINK_FREQ_400MHZ / 100000;
	ret = vsc_acquire_camera_sensor(&ov01a1s->conf, NULL, NULL, &ov01a1s->status);
	if (!ret) {
		ov01a1s->power_type = OV01A1S_USE_INTEL_VSC;
		return 0;
	} else if (ret != -EAGAIN) {
		return ret;
	}
#endif
#if IS_ENABLED(CONFIG_INTEL_SKL_INT3472)
	ret = ov01a1s_parse_gpio(ov01a1s);
#elif IS_ENABLED(CONFIG_POWER_CTRL_LOGIC)
	ret = power_ctrl_logic_set_power(1);
#endif
#if IS_ENABLED(CONFIG_INTEL_SKL_INT3472) || IS_ENABLED(CONFIG_POWER_CTRL_LOGIC)
	if (!ret) {
		ov01a1s->power_type = OV01A1S_USE_INT3472;
		return 0;
	}
#endif
	if (ret == -EAGAIN)
		return -EPROBE_DEFER;

	return ret;
}

static int ov01a1s_probe(struct i2c_client *client)
{
	struct ov01a1s *ov01a1s;
	int ret = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0) && \
    IS_ENABLED(CONFIG_INTEL_VSC)
	ret = init_vsc_symbols();
	if (ret)
		return ret;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 8, 0)
	if (!dmi_check_system(skip_hwcfg_check_dmi_table)) {
#endif
		/* Check HW config */
		ret = ov01a1s_check_hwcfg(&client->dev);
		if (ret) {
			dev_err(&client->dev, "failed to check hwcfg: %d", ret);
			return ret;
		}
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 8, 0)
	}
#endif

	ov01a1s = devm_kzalloc(&client->dev, sizeof(*ov01a1s), GFP_KERNEL);
	if (!ov01a1s)
		return -ENOMEM;

	ov01a1s->client = client;
	ret = ov01a1s_parse_power(ov01a1s);
	if (ret)
		return ret;

	v4l2_i2c_subdev_init(&ov01a1s->sd, client, &ov01a1s_subdev_ops);
#if IS_ENABLED(CONFIG_INTEL_SKL_INT3472)
	/* In other cases, power is up in ov01a1s_parse_power */
	if (ov01a1s->power_type == OV01A1S_USE_INT3472)
		ov01a1s_power_on(&client->dev);
#endif
	ret = ov01a1s_identify_module(ov01a1s);
	if (ret) {
		dev_err(&client->dev, "failed to find sensor: %d", ret);
		goto probe_error_power_off;
	}

	mutex_init(&ov01a1s->mutex);
	ov01a1s->cur_mode = &supported_modes[0];
	ret = ov01a1s_init_controls(ov01a1s);
	if (ret) {
		dev_err(&client->dev, "failed to init controls: %d", ret);
		goto probe_error_v4l2_ctrl_handler_free;
	}

	ov01a1s->sd.internal_ops = &ov01a1s_internal_ops;
	ov01a1s->sd.flags |= V4L2_SUBDEV_FL_HAS_DEVNODE;
	ov01a1s->sd.entity.ops = &ov01a1s_subdev_entity_ops;
	ov01a1s->sd.entity.function = MEDIA_ENT_F_CAM_SENSOR;
	ov01a1s->pad.flags = MEDIA_PAD_FL_SOURCE;
	ret = media_entity_pads_init(&ov01a1s->sd.entity, 1, &ov01a1s->pad);
	if (ret) {
		dev_err(&client->dev, "failed to init entity pads: %d", ret);
		goto probe_error_v4l2_ctrl_handler_free;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 13, 0)
	ret = v4l2_async_register_subdev_sensor_common(&ov01a1s->sd);
#else
	ret = v4l2_async_register_subdev_sensor(&ov01a1s->sd);
#endif
	if (ret < 0) {
		dev_err(&client->dev, "failed to register V4L2 subdev: %d",
			ret);
		goto probe_error_media_entity_cleanup;
	}

	/*
	 * Device is already turned on by i2c-core with ACPI domain PM.
	 * Enable runtime PM and turn off the device.
	 */
	pm_runtime_set_active(&client->dev);
	pm_runtime_enable(&client->dev);
	pm_runtime_idle(&client->dev);

	return 0;

probe_error_media_entity_cleanup:
	media_entity_cleanup(&ov01a1s->sd.entity);

probe_error_v4l2_ctrl_handler_free:
	v4l2_ctrl_handler_free(ov01a1s->sd.ctrl_handler);
	mutex_destroy(&ov01a1s->mutex);

probe_error_power_off:
	ov01a1s_power_off(&client->dev);

	return ret;
}

static const struct dev_pm_ops ov01a1s_pm_ops = {
	SET_SYSTEM_SLEEP_PM_OPS(ov01a1s_suspend, ov01a1s_resume)
	SET_RUNTIME_PM_OPS(ov01a1s_power_off, ov01a1s_power_on, NULL)
};

#ifdef CONFIG_ACPI
static const struct acpi_device_id ov01a1s_acpi_ids[] = {
	{ "OVTI01AS" },
	{}
};

MODULE_DEVICE_TABLE(acpi, ov01a1s_acpi_ids);
#endif

static struct i2c_driver ov01a1s_i2c_driver = {
	.driver = {
		.name = "ov01a1s",
		.pm = &ov01a1s_pm_ops,
		.acpi_match_table = ACPI_PTR(ov01a1s_acpi_ids),
	},
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
	.probe_new = ov01a1s_probe,
#else
	.probe = ov01a1s_probe,
#endif
	.remove = ov01a1s_remove,
};

module_i2c_driver(ov01a1s_i2c_driver);

MODULE_AUTHOR("Xu, Chongyang <chongyang.xu@intel.com>");
MODULE_AUTHOR("Lai, Jim <jim.lai@intel.com>");
MODULE_AUTHOR("Qiu, Tianshu <tian.shu.qiu@intel.com>");
MODULE_AUTHOR("Shawn Tu <shawnx.tu@intel.com>");
MODULE_AUTHOR("Bingbu Cao <bingbu.cao@intel.com>");
MODULE_DESCRIPTION("OmniVision OV01A1S sensor driver");
MODULE_LICENSE("GPL v2");
